(*************************************************************
 *                Auto-assemblage de graphes                 *
 *************************************************************
   Auteur  : Fabien TARISSAN
   Fichier : move.ml
   Date    : 09/06/04
 *************************************************************
   Description : To compute the next move of a robot
 *************************************************************)

open Param;;
open Agent;;
open System;;

open Eval;;


let alarme_list = ref [];;

(* INITIALISATION *)
(******************)

let init_compute_move = List.map 
  (fun a -> (
     a, ref a.span, 
     ref a.cycle, 
     (ref 0., ref 0.), 
     (ref 0., ref 0.), 
     (ref 0., ref 0.) 
   )
  )
;;


(* TESTS DES EFFETS *)
(********************)

let neg_couple (x,y) = (-.x, -.y)
;;

let compute_rep (x,y) dist = 
  if dist = 0.
  then (10. , 10.)
  else  if dist >= dist_perc
  then (0.,0.)
  else (-.x *. 1. /. dist, -.y *. 1. /. dist)
;;

let add (x,y) (tmp_x,tmp_y) =
  tmp_x := x +. !tmp_x;
  tmp_y := y +. !tmp_y
;;


let attract_or_repulse (x,y) dist =
  let d = dist /. 2. -. ecart_equi in
    (d *. x, d *. y)
;;


let vitesse a =  sqrt (a.dx*.a.dx +. a.dy*.a.dy)
;;

let positive i = if i < 0. then -.i else i;;
let negative i = if i < 0. then i else -.i;;

let float_circle = float size_circle;;
let float_window_x = float size_window_x;;
let float_window_y = float size_window_y;;

let norm (x,y) = 
  if x = 0. && y = 0. then (1.,1.) else
    if x < 0. || (x = 0. && y > 0.)
    then (-. x, -. y)
    else (x,y)
;;

(* FIN *)
(*******)

let modif_vitesse a =
  let v =
    if a.span != [] 
    then vitesse_con_min
    else vitesse_rech_min
  in
    if vitesse a < v
    then
      let (xx,yy) = (Random.float 1., Random.float 1.) in
      let p = Random.int 2 in
	if p = 0
	then a.dx <- a.dx *. vitesse_up +. 1.1 *. xx
	else a.dx <- a.dx *. vitesse_up -. 1.1 *. xx;
	let p = Random.int 2 in
	  if p = 0
	  then a.dy <- a.dy *. vitesse_up +. 1.1 *. yy 
	  else a.dy <- a.dy *. vitesse_up -. 1.1 *. yy;
	  true;
    else false
;;



let finish_compute_move l =

  let verif_rebond a1 (a2,_,_,_,_,_) =
    let (x,y) = (a2.pos_x -. a1.pos_x, a2.pos_y -. a1.pos_y) in
    let (vx,vy) = norm (y , -. x) in
    let dist = sqrt (vx*.vx +. vy*.vy) in
      if dist < 3. *. float_circle
      then (
	let costh = vx/.dist in
	let th = acos costh in
	let th2 = 2. *. th in
	let cos2th = cos th2 in
	let sin2th = sin th2 in
	  
	let calc a (dirx, diry)=
	  if dirx *. a.dx +. diry *. a.dy > 0. then
	    let (x,y) = (a.dx *. cos2th -. a.dy *. sin2th,
			 -. a.dy *. cos2th -. a.dx *. sin2th) 
	    in
	      a.dx <- x;
	      a.dy <- y
	in

	  calc a1 (x,y);
	  calc a2 (-. x, -. y)
      )
	
  in
    
  let rec final l =
    match l with
	[] -> ()
      | (hd,_,_,_,_,_)::tl ->
	  List.iter (verif_rebond hd) tl;

	  if (hd.pos_x -. 2. *. float_circle) <= 0.  
	  then hd.dx <- positive hd.dx
	  else if (hd.pos_x +. 2. *. float_circle) >= float_window_x 
	  then hd.dx <- negative hd.dx;
	  
	  if (hd.pos_y -. 2. *. float_circle) <= 0.  
	  then hd.dy <- positive hd.dy
	  else if (hd.pos_y +. 2. *. float_circle) >= float_window_y 
	  then hd.dy <- negative hd.dy;
	  
	  hd.pos_x <- hd.pos_x +. hd.dx;
	  hd.pos_y <- hd.pos_y +. hd.dy;
	  final tl
  in
    
  let calcul_depl (a,_,_, (sp_x,sp_y) , (cy_x, cy_y) , (rep_x, rep_y)) =
    let x = coeff_join *. !sp_x 
      +. coeff_self *. !cy_x
      +. coeff_rep *. !rep_x
    in
    let y =  coeff_join *. !sp_y
      +. coeff_self *. !cy_y
      +. coeff_rep *. !rep_y
    in
      a.dx <- x +. a.dx *. coeff_frot ;
      a.dy <- y +. a.dy *. coeff_frot ;
      while(modif_vitesse a) do (); done;

  in

    List.iter calcul_depl l;
    final l
;;



let recup_agent a l =
  let rec recup_rec acc l =
    match l with
	[] -> None
      | hd::tl ->
          if a == hd 
	  then Some(acc@tl)
          else recup_rec (hd::acc) tl
  in
    recup_rec [] l
;;


let translate_alarme liste sys =
  List.iter (fun a -> 
	       a.span <- [];
	       a.cycle <- [];
	       a.state <- Alarme(Sys.time ())
            ) liste;
  let f a =
    match recup_agent a sys.active with
	None -> (
	  match recup_agent a sys.update with
	      None -> (
		match recup_agent a sys.passive with
		    None -> ()
		  | Some(l) -> sys.passive <- l; 
		      sys.break <- a::sys.break
	      )
	    | Some(l) -> sys.update <- l; sys.break <- a::sys.break
	)
      | Some(l) -> sys.active <- l; sys.break <- a::sys.break
  in
    List.iter f liste
;;
    


(* TESTS D'APPARTENANCE *)
(************************)

let mem_and_extract_com com liste =
  let rec mem_rec acc l =
    match l with
	[] -> false
      | hd::tl ->
	  if hd = com
	  then (
	    liste := acc@tl;
	    true
	  )
	  else mem_rec (hd::acc) tl
  in
    mem_rec [] !liste
;;

let exist_connection l1 l2 =
  let rec exist_rec acc l =
    match l with
	[] -> false
      | hd::tl ->
	  if mem_and_extract_com hd l1
	  then (
	    l2 := acc@tl;
	    true
	  )
	  else exist_rec (hd::acc) tl
  in
    exist_rec [] !l2
;;


(* CALCUL *)
(**********)


let compute_neigh (a, sp, cy, vs, vc, vrep) liste =
  let rec compute_connection l =
    match l with
	[] -> ()
      | (rob,rob_sp, rob_cy, rob_vs, rob_vc, rob_vrep)::tl ->
	  let (vect_x,vect_y) as vect = 
	    (rob.pos_x -. a.pos_x , rob.pos_y -. a.pos_y) 
	  in let dist = sqrt (vect_x*.vect_x +. vect_y*.vect_y) 
	  in let rep =  compute_rep vect dist in
	    add rep vrep;
	    add (neg_couple rep) rob_vrep;
	    
	    (
	      if exist_connection sp rob_sp 
	      then (
		if dist < link_max then
		  let vect = attract_or_repulse vect dist in
		    add vect vs;
		    add (neg_couple vect) rob_vs
		else alarme_list := a::(rob::!alarme_list)
		  
	      )
	      else if exist_connection cy rob_cy
	      then (
		if dist < link_max then
		  let vect = attract_or_repulse vect dist in
		    add vect vc;
		    add (neg_couple vect) rob_vc
		else alarme_list := a::(rob::!alarme_list)
	      )
	    );
	    
	    compute_connection tl
	      
  in
    compute_connection liste;
    if !sp != [] || !cy != [] then alarme_list := a::!alarme_list
;;



(* Calcul de la prochaine localisation de chaque agent *)

let compute_next_move sys =
  alarme_list := [];
  let l = init_compute_move 
    (sys.active@sys.update@sys.passive@sys.break@sys.final) 
  in let rec compute_rec l =
      match l with
	  [] -> ()
	| hd::tl ->
	    compute_neigh hd tl;
	    compute_rec tl;
  in
    compute_rec l;
    finish_compute_move l;
    translate_alarme !alarme_list sys    
;;
