(*************************************************************
 *              Auto-assemblage de graphes                   *
 *************************************************************
   Auteur  : Fabien TARISSAN
   Fichier : agent.ml
   Date    : 09/06/04
 *************************************************************
   Description : Syntaxe abstraite pour les agents
 *************************************************************)

exception Agent_Error of string;;

(* Type des connexions *)
type com = string
;;

(* Cration d'un nouvelle connexion *)
let create_com =
  let str = ref "x" and c = ref 0 in 
    function () -> 
      if !c = max_int
      then ( 
	str := (!str)^"x";
	c := 0
      );
      c := !c+1;
      (!str)^(string_of_int(!c))
;;

(* Type de la reprsentation des graphes *)
type graph = int * int
;;

(* Type utiliser pour propager les informations *)
(* sur la structure de la nouvelle composante   *)
type shift = int array
;;

(* Type de la reprsentation du role d'un agent *)
type role = int 
;;

(* Type de l'tat d'un agent *)
type state = 
    Passive 
  | Active of graph
  | Update of com list * shift * bool * graph option 
  | Alarme of float
  | Final
;;


(* Type des agents *)
type agent = {
  mutable ident : com;      (* identifiant de groupe                *)
  mutable span  : com list; (* voisins dans l'arbre de recouvrement *)
  mutable cycle : com list; (* autres voisins                       *)
  mutable role  : role;     (* rle jou par dans la composante     *)
  mutable state : state;    (* tat (type) de l'agent               *)

  mutable pos_x : float;    (* positions de l'agent *)
  mutable pos_y : float; 

  mutable dx : float;       (* dernier dplacement de l'agent *)
  mutable dy : float
}
;;


(* Pour rcuprer l'image de la composante *)
(* connexe par le biai d'un agent actif    *)

let recup_image a =
  match a.state with
      Active(g) -> g
    | _ -> raise (Agent_Error "recup_image : Agent non actif")
;;


let recup_alarme a =
  match a.state with
      Alarme(t) -> t
    | _ -> raise( Agent_Error "Agent not in alarme mode")
;;
